@@ -8,17 +8,18 @@ require overload;
 # array based objects. This may not remain the case.
 
 
-use Scalar::Util qw/reftype weaken/;
+use Scalar::Util qw/reftype/;
 
 our $AUTOLOAD;
 
-sub new {
-  my $proto = shift;
-  my $self = {};
-  bless $self, ref($proto) || $proto;
-}
+our $VERSION="2.06";
 
-sub make_proxy {
+use Pixie::Object;
+use Pixie::FinalMethods;
+use Pixie::Complicity;
+use base 'Pixie::Object';
+
+sub px_make_proxy {
   my $self = shift;
   my($oid, $obj) = @_;
   my $proxied_class = ref($obj);
@@ -26,85 +27,114 @@ sub make_proxy {
 
   $real_class .= '::Overloaded' if overload::Overloaded($proxied_class);
   $real_class->new->_oid($oid)
-                  ->class($proxied_class);
+                  ->px_class($proxied_class);
 }
 
-sub restore {
-  my $class = $_[0]->class;
-  my $pixie = $_[0]->the_store;
-  $_[0]->clear_the_store;
-  my $real_obj = $_[0]->fetch_from($pixie);
+sub px_restore {
+  my $class = $_[0]->px_class;
+  my $pixie = $_[0]->px_the_store;
+  $_[0]->px_clear_the_store;
+  my $real_obj = $_[0]->px_fetch_from($pixie);
   return $_[0] = undef unless defined $real_obj;
-  $_[0]->populate_from(bless $real_obj, 'Class::Whitehole');
-  bless $_[0], $class;
+  $_[0]->populate_from($real_obj);
+  bless $real_obj, 'Class::Whitehole';
+  my $ret = bless $_[0], $class;
 }
 
-sub fetch_from {
+sub px_fetch_from {
   my $self = shift;
   my $pixie = shift;
 
-  local $pixie->cache->{$self->_oid};
-  $pixie->get($self->_oid);
+  my $oid = $self->_oid;
+
+  $pixie->get_with_strategy($oid, $self->px_lock_strategy);
 }
 
 sub isa {
   my $self = shift;
   my($class) = @_;
-  $self->UNVERSAL::isa($class) || $self->class->isa($class);
+  $self->UNIVERSAL::isa($class) || ref($self) && $self->px_class->isa($class);
 }
 
 sub can {
   my $self = shift;
   my($method) = @_;
 
-  $self->UNVERSAL::can($method) ||
-      ref($self) && $self->restore->can($method);
+  $self->UNIVERSAL::can($method) ||
+      ref($self) && $self->px_restore->can($method);
 }
 
 
-{
-  my %store_hash;
-  sub the_store {
-    my $self = shift;
-    my $key = $self->_oid;
-    if (@_) {
-      $store_hash{$key} = shift;
-      weaken $store_hash{$key};
-      return $self;
-    }
-    else {
-      return $store_hash{$key};
-    }
-  }
-
-  sub clear_the_store {
-    my $self = shift;
-    delete $store_hash{$self->_oid};
-  }
-}
-
 sub STORABLE_freeze {
   my $self = shift;
   my $cloning = shift;
   return if $cloning;
 
-  return $self->_oid, [$self->class];
+  return $self->_oid, [$self->px_class];
 }
 
 sub STORABLE_thaw {
   my($target, $cloning, $oid, $class) = @_;
   return if $cloning;
   $target->_oid($oid);
-  $target->class($class->[0]);
+  $target->px_class($class->[0]);
   return $target;
 }
 
-sub DESTROY { }
+sub _px_insertion_thaw {
+  my $self = shift;
+  $self->px_the_store(Pixie->get_the_current_pixie);
+  return $self;
+}
+
+sub _px_insertion_freeze {
+  my $self = shift;
+  my $dupe = ref($self)->new->_oid($self->_oid)
+                            ->px_class($self->px_class);
+}
+
+
+
+sub _px_extraction_thaw {
+  my $self = shift;
+  my $pixie = Pixie->get_the_current_pixie($self->_oid);
+  my $ret = Pixie->get_the_current_pixie->cache_get($self->_oid);
+  if ( defined $ret ) {
+    bless $self, 'Class::Whitehole';
+    $pixie->forget_about($self);
+    return $ret;
+  }
+
+  $self->px_lock_strategy( $pixie->get_the_current_lock_strategy ||
+                           $pixie->lock_strategy );
+
+
+  if ($self->px_class->px_is_immediate) {
+    my $oid = $self->_oid;
+    bless $self, 'Class::Whitehole';
+    Pixie->get_the_current_pixie->_get($oid);
+  }
+  else {
+    $self->px_the_store($pixie);
+    $pixie->cache_insert($self);
+    return $self;
+  }
+}
+
+sub DESTROY {
+  my $self = shift;
+  local $@ = $@;
+  return unless ref $self;
+  my $store = $self->px_the_store;
+  if (defined $store) {
+    $store->forget_about($self);
+  }
+}
 
 sub AUTOLOAD {
   my $method = $AUTOLOAD;
   $method =~ s/.*:://;
-  $_[0]->restore->$method(@_[1..$#_]);
+  $_[0]->px_restore->$method(@_[1..$#_]);
 }
 
 package Pixie::Proxy::ARRAY;
@@ -127,7 +157,7 @@ sub _oid {
   }
 }
 
-sub class {
+sub px_class {
   my $self = shift;
   if (@_) {
     $self->[1] = shift;
@@ -143,6 +173,35 @@ sub populate_from {
   return $_[0];
 }
 
+sub px_the_store {
+  my $self = shift;
+  if (@_) {
+    $self->[2] = shift;
+    return $self;
+  }
+  else {
+    return $self->[2];
+  }
+}
+
+sub px_clear_the_store {
+  my $self = shift;
+  $self->[2] = undef;
+  return $self;
+}
+
+sub px_lock_strategy {
+  my $self = shift;
+  if (@_) {
+    $self->[3] = shift;
+    return $self;
+  }
+  else {
+    return $self->[3];
+  }
+}
+
+
 package Pixie::Proxy::HASH;
 
 use base 'Pixie::Proxy';
@@ -164,13 +223,20 @@ sub _oid {
   }
 }
 
-sub class {
+
+sub px_oid { $_[0]->_oid }
+
+sub px_class {
   my $self = shift;
 
   if (@_) {
     $self->{class} = shift;
     return $self;
   } else {
+    unless (ref($self)) {
+      require Carp;
+      Carp::confess "Invalid thing: $self";
+    }
     return $self->{class};
   }
 }
@@ -183,6 +249,33 @@ sub populate_from {
   return $_[0];
 }
 
+sub px_the_store {
+  my $self = shift;
+  if (@_) {
+    $self->{_the_store} = shift;
+    return $self;
+  }
+  else {
+    return $self->{_the_store};
+  }
+}
+
+sub px_clear_the_store {
+  my $self = shift;
+  delete $self->{_the_store};
+}
+
+sub px_lock_strategy {
+  my $self = shift;
+  if (@_) {
+    $self->{_lock_strategy} = shift;
+    return $self;
+  }
+  else {
+    return $self->{_lock_strategy};
+  }
+}
+
 package Pixie::Proxy::Overloaded;
 
 my %FALLBACK = ( '!' => \&bool_not,
@@ -198,10 +291,10 @@ use overload
     nomethod => sub {
       no strict 'refs';
       my $method = pop;
-      my $class = $_[0]->class;
+      my $class = $_[0]->px_class;
       my $fb = $ {$class . "::()"};
       if ( my $sub = overload::ov_method( overload::mycan($class, "\($method"), $class) ) {
-        $_[0]->restore;
+        $_[0]->px_restore;
         &$sub;
       }
       elsif (!defined($fb) || $fb) {
@@ -228,7 +321,7 @@ sub bool_not {
     return;
   }
   else {
-    $_[0]->restore;
+    $_[0]->px_restore;
     return ! $_[0];
   }
 }
@@ -238,7 +331,7 @@ sub bool {
     return 1;
   }
   else {
-    $_[0]->restore;
+    $_[0]->px_restore;
     return $_[0];
   }
 }
